library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity REORDER is
  port (
    RESET     : in  std_logic;
    CLK       : in  std_logic;
    ROHEAD    : in  std_logic;                      -- <1,1,u>
    ROIN_I    : in  std_logic_vector(13 downto 0);  -- <14,6,t>
    ROIN_Q    : in  std_logic_vector(13 downto 0);  -- <14,6,t>
    ROOUTHEAD : out std_logic;                      -- <1,1,u>
    ROOUT_I   : out std_logic_vector(13 downto 0);  -- <14,6,t>
    ROOUT_Q   : out std_logic_vector(13 downto 0)); -- <14,6,t>
end;

architecture RTL of REORDER is

signal COUNT          : std_logic_vector(5 downto 0);
signal BANK           : std_logic;                      -- Input bank select signal
signal BANKEN         : std_logic;
type shiftreg64 is array (0 to 63) of std_logic_vector (13 downto 0); -- <14,6,t> 
signal REG0_I, REG0_Q : shiftreg64;
signal REG1_I, REG1_Q : shiftreg64;
signal REGIN_I, REGIN_Q :std_logic_vector (13 downto 0); -- <14,6,t>

begin
    
-- 64 CYCLE COUNTER
    process ( CLK ) begin 
      if (RESET = '1') then 
         COUNT <="000000";
         BANKEN <= '0';
      elsif ( CLK'event and CLK = '1' ) then
          if (ROHEAD = '1') then 
            COUNT <= "000000";
            BANKEN <= '1';           
          elsif (COUNT = "111111") then
            BANKEN <= '0';
          else
            COUNT <= unsigned(COUNT) + '1';
          end if;
      end if; 
    end process;
    
-- BANK SELECT GENERATION
    process ( CLK ) begin
      if (RESET = '1') then 
        BANK <= '0';
      elsif ( CLK'event and CLK = '1' ) then
        if (ROHEAD = '1') then 
            BANK <= not BANK;
        end if;
      end if; 
  end process;
  
-- INPUT REGISTER
   process (CLK) begin
       if (CLK'event and CLK = '1' ) then
           REGIN_I <= ROIN_I;
           REGIN_Q <= ROIN_Q;
       end if;
   end process;

-- SHIFT REISTER
    process (CLK) begin
        if (CLK'event and CLK = '1' ) then
            if (BANKEN = '1')then
               if (BANK = '0') then
                --
                -- BANK 0 SERIAL INPUT
                --
                for I in 0 to 62 loop
                    REG0_I(I) <= REG0_I(I+1);
                    REG0_Q(I) <= REG0_Q(I+1);
                end loop;
                    REG0_I(63) <= REGIN_I;
                    REG0_Q(63) <= REGIN_Q;
                --
                -- BANK 1 REORDER OUPUT
                --
                for I in 0 to 3 loop
                    REG1_I(I+ 0) <= REG1_I(I+16);
                    REG1_I(I+16) <= REG1_I(I+32);
                    REG1_I(I+32) <= REG1_I(I+48);
                    REG1_I(I+48) <= REG1_I(I+ 4);
                    REG1_I(I+ 4) <= REG1_I(I+20);
                    REG1_I(I+20) <= REG1_I(I+36);
                    REG1_I(I+36) <= REG1_I(I+52);
                    REG1_I(I+52) <= REG1_I(I+ 8);
                    REG1_I(I+ 8) <= REG1_I(I+24);
                    REG1_I(I+24) <= REG1_I(I+40);
                    REG1_I(I+40) <= REG1_I(I+56);
                    REG1_I(I+56) <= REG1_I(I+12);
                    REG1_I(I+12) <= REG1_I(I+28);
                    REG1_I(I+28) <= REG1_I(I+44);
                    REG1_I(I+44) <= REG1_I(I+60);
                    REG1_Q(I+ 0) <= REG1_Q(I+16);
                    REG1_Q(I+16) <= REG1_Q(I+32);
                    REG1_Q(I+32) <= REG1_Q(I+48);
                    REG1_Q(I+48) <= REG1_Q(I+ 4);
                    REG1_Q(I+ 4) <= REG1_Q(I+20);
                    REG1_Q(I+20) <= REG1_Q(I+36);
                    REG1_Q(I+36) <= REG1_Q(I+52);
                    REG1_Q(I+52) <= REG1_Q(I+ 8);
                    REG1_Q(I+ 8) <= REG1_Q(I+24);
                    REG1_Q(I+24) <= REG1_Q(I+40);
                    REG1_Q(I+40) <= REG1_Q(I+56);
                    REG1_Q(I+56) <= REG1_Q(I+12);
                    REG1_Q(I+12) <= REG1_Q(I+28);
                    REG1_Q(I+28) <= REG1_Q(I+44);
                    REG1_Q(I+44) <= REG1_Q(I+60);
                end loop;
                    REG1_I(60)   <= REG1_I(1);
                    REG1_I(61)   <= REG1_I(2);
                    REG1_I(62)   <= REG1_I(3);
                    REG1_Q(60)   <= REG1_Q(1);
                    REG1_Q(61)   <= REG1_Q(2);
                    REG1_Q(62)   <= REG1_Q(3);
                else
                --
                -- BANK 1 SERIAL INPUT
                --
                for I in 0 to 62 loop
                    REG1_I(I) <= REG1_I(I+1);
                    REG1_Q(I) <= REG1_Q(I+1);
                end loop;
                    REG1_I(63) <= REGIN_I;
                    REG1_Q(63) <= REGIN_Q;
                --
                -- BANK 0 REORDER OUPUT
                --
                for I in 0 to 3 loop
                    REG0_I(I+ 0) <= REG0_I(I+16);
                    REG0_I(I+16) <= REG0_I(I+32);
                    REG0_I(I+32) <= REG0_I(I+48);
                    REG0_I(I+48) <= REG0_I(I+ 4);
                    REG0_I(I+ 4) <= REG0_I(I+20);
                    REG0_I(I+20) <= REG0_I(I+36);
                    REG0_I(I+36) <= REG0_I(I+52);
                    REG0_I(I+52) <= REG0_I(I+ 8);
                    REG0_I(I+ 8) <= REG0_I(I+24);
                    REG0_I(I+24) <= REG0_I(I+40);
                    REG0_I(I+40) <= REG0_I(I+56);
                    REG0_I(I+56) <= REG0_I(I+12);
                    REG0_I(I+12) <= REG0_I(I+28);
                    REG0_I(I+28) <= REG0_I(I+44);
                    REG0_I(I+44) <= REG0_I(I+60);
                    REG0_Q(I+ 0) <= REG0_Q(I+16);
                    REG0_Q(I+16) <= REG0_Q(I+32);
                    REG0_Q(I+32) <= REG0_Q(I+48);
                    REG0_Q(I+48) <= REG0_Q(I+ 4);
                    REG0_Q(I+ 4) <= REG0_Q(I+20);
                    REG0_Q(I+20) <= REG0_Q(I+36);
                    REG0_Q(I+36) <= REG0_Q(I+52);
                    REG0_Q(I+52) <= REG0_Q(I+ 8);
                    REG0_Q(I+ 8) <= REG0_Q(I+24);
                    REG0_Q(I+24) <= REG0_Q(I+40);
                    REG0_Q(I+40) <= REG0_Q(I+56);
                    REG0_Q(I+56) <= REG0_Q(I+12);
                    REG0_Q(I+12) <= REG0_Q(I+28);
                    REG0_Q(I+28) <= REG0_Q(I+44);
                    REG0_Q(I+44) <= REG0_Q(I+60);
                end loop;
                    REG0_I(60)   <= REG0_I(1);
                    REG0_I(61)   <= REG0_I(2);
                    REG0_I(62)   <= REG0_I(3);
                    REG0_Q(60)   <= REG0_Q(1);
                    REG0_Q(61)   <= REG0_Q(2);
                    REG0_Q(62)   <= REG0_Q(3);
               end if;    
           end if;
       end if;
    end process;
    
-- OUTHEAD GENERATOR
   OH1: process (CLK) begin
        if (RESET = '1') then 
         ROOUTHEAD <= '0';
        elsif (CLK'event and CLK = '1' ) then
           if (ROHEAD = '1') then 
            ROOUTHEAD <= '1';
           else
            ROOUTHEAD <= '0';
          end if;
        end if;
   end process;

-- OUTPUT SELECTOR
ROOUT_I <= REG1_I(0) when BANK = '0' else
           REG0_I(0);
ROOUT_Q <= REG1_Q(0) when BANK = '0' else
           REG0_Q(0);

end;
