library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity STAGE2 is
  port (
    RESET     : in  std_logic;
    CLK       : in  std_logic;
    S2HEAD    : in  std_logic;                      -- <1,1,u>
    S2IN_I    : in  std_logic_vector(13 downto 0);  -- <14,6,t>
    S2IN_Q    : in  std_logic_vector(13 downto 0);  -- <14,6,t>
    S2OUTHEAD : out std_logic;                      -- <1,1,u>
    S2OUT_I   : out std_logic_vector(13 downto 0);  -- <14,6,t>
    S2OUT_Q   : out std_logic_vector(13 downto 0)); -- <14,6,t>
end;

architecture RTL of STAGE2 is

component CMULT is
  port(AIN_I     : in  std_logic_vector(13 downto 0);  -- <14,6,t>
       AIN_Q     : in  std_logic_vector(13 downto 0);  -- <14,6,t>
       BIN_I     : in  std_logic_vector(9  downto 0);  -- <10,0,t>
       BIN_Q     : in  std_logic_vector(9  downto 0);  -- <10,0,t>
       YOUT_I    : out std_logic_vector(13 downto 0);  -- <14,6,t>
       YOUT_Q    : out std_logic_vector(13 downto 0)); -- <14,6,t>   
end component;

component TWIDDLE
  port (
    ADDR : in  std_logic_vector(5 downto 0); -- <6,6,u>
    W_I  : out std_logic_vector(9 downto 0); -- <10,0,t>
    W_Q  : out std_logic_vector(9 downto 0)); -- <10,0,t>
end component;

signal  HEAD  : std_logic_vector(16 downto 0);  -- 17 FFs for Head Signal Delay
signal  COUNT : std_logic_vector(5 downto 0);  -- count 0 to 63 cycles
signal  PHASE : std_logic_vector(1 downto 0);  -- PHASE 0 to 3
signal  TF    : std_logic_vector(5 downto 0);  -- TWIDDLE FACTOR INDEX
signal  S1, S2, S3 : std_logic;

signal REGIN_I : std_logic_vector (13 downto 0); -- <14,6,t>
signal REGIN_Q : std_logic_vector (13 downto 0); -- <14,6,t>
type shiftreg28 is array (0 to 27) of std_logic_vector (13 downto 0); -- <14,6,t> 
signal REG_I  : shiftreg28;
signal REG_Q  : shiftreg28;
signal A_I, A_Q : std_logic_vector (13 downto 0); -- <14,6,t> 
signal B_I, B_Q : std_logic_vector (13 downto 0); -- <14,6,t> 
signal C_I, C_Q : std_logic_vector (13 downto 0); -- <14,6,t> 
signal D_I, D_Q : std_logic_vector (13 downto 0); -- <14,6,t> 
signal Y_I, Y_Q : std_logic_vector (13 downto 0); -- <14,6,t>  butterfly output
signal W_I, W_Q : std_logic_vector(9 downto 0); -- <10,0,t>

begin

-- OUTHEAD GENERATION
    process ( CLK, RESET ) begin 
      if (RESET = '1') then
          for I in 0 to 16 loop
             HEAD(I) <= '0';
         end loop;
      elsif ( CLK'event and CLK = '1' ) then
            HEAD(0) <= S2HEAD;
            for I in 1 to 16 loop
               HEAD(I) <= HEAD(I-1);
           end loop;
      end if; 
  end process;
    S2OUTHEAD <= HEAD(16);

-- 64 CYCLE COUNTER
    process ( CLK, RESET ) begin 
    if (RESET = '1') then
        COUNT <= "000000";
    elsif ( CLK'event and CLK = '1' ) then
        if (S2HEAD = '1') then 
            COUNT <= "000000"; 
        else
            COUNT <= unsigned(COUNT) + '1';
        end if;
    end if; 
    end process;

-- PHASE GENERATOR
    PHASE <= COUNT(3 downto 2);  --henkou

-- TWIDDLE FACTOR GENERATOR
    TF <= unsigned(COUNT(1 downto 0)) * (unsigned(PHASE(1 downto 0)) & "00");  --henkou

-- SELECT SIGNAL GENERATOR
    S1 <= PHASE(0) or PHASE(1);
    S2 <= PHASE(1);
    S3 <= PHASE(0) and PHASE(1);

-- SHIFT REGISTER 
    process ( CLK ) begin 
      if ( CLK'event and CLK = '1' ) then
        REGIN_I  <=  S2IN_I;    REGIN_Q  <=  S2IN_Q;
        REG_I(0) <=  REGIN_I;   REG_Q(0) <=  REGIN_Q; 
        for I in 1 to 27 loop
           REG_I(I) <=  REG_I(I-1);  REG_Q(I) <=  REG_Q(I-1);
       end loop;
      end if; 
    end process;

-- A,B,C,D GENERATION
    A_I <= REG_I(15);
    A_Q <= REG_Q(15);
    B_I <= REG_I(3) when S1 = '0' else REG_I(19);
    B_Q <= REG_Q(3) when S1 = '0' else REG_Q(19);
    C_I <= REG_I(7) when S2 = '0' else REG_I(23);
    C_Q <= REG_Q(7) when S2 = '0' else REG_Q(23);
    D_I <= REG_I(11) when S3 = '0' else REG_I(27);
    D_Q <= REG_Q(11) when S3 = '0' else REG_Q(27);


-- RADIX-4 BUTTERFLY 

process (PHASE, A_I, B_I, C_I, D_I, A_Q, B_Q, C_Q, D_Q) begin
  case PHASE is
    when "00" => 
       Y_I <= signed(A_I) + signed(B_I) + signed(C_I) + signed (D_I);
       Y_Q <= signed(A_Q) + signed(B_Q) + signed(C_Q) + signed (D_Q);
    when "01" => 
       Y_I <= signed(A_Q) + signed(B_I) - signed(C_Q) - signed (D_I);
       Y_Q <= - signed(A_I) + signed(B_Q) + signed(C_I) - signed (D_Q);
    when "10" => 
       Y_I <= signed(A_I) - signed(B_I) + signed(C_I) - signed (D_I);
       Y_Q <= signed(A_Q) - signed(B_Q) + signed(C_Q) - signed (D_Q);
    when "11" => 
       Y_I <= signed(A_Q) - signed(B_I) - signed(C_Q) + signed (D_I);
       Y_Q <= - signed(A_I) - signed(B_Q) + signed(C_I) + signed (D_Q);
    when others => 
       Y_I <= (others => 'X');
       Y_Q <= (others => 'X');
  end case;
end process;

TW0: TWIDDLE port map (TF, W_I, W_Q);

CM0: CMULT port map (Y_I, Y_Q, W_I, W_Q, S2OUT_I, S2OUT_Q);

end;
